//*****************************************************************************
// USB read/write utility for FTDI FT245 series.
// Tested on Cygwin g++
//*****************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <time.h>
#include "ftd2xx.h"

#define N_ELEMS(A) (sizeof(A)/sizeof((A)[0]))

//*****************************************************************************
// Declare USB procedures. This part is application independent.
//*****************************************************************************

static const char* getFtErrorMessage(FT_STATUS ftStatus);

// Enumerate procedure informations as triples
// (Procedure name, param types, param names). 

#define USB_PROCS \
  \
PROC(FT_Open, \
     (int iDevice, \
      FT_HANDLE *pftHandle), \
     (iDevice, pftHandle)) \
PROC(FT_ListDevices, \
     (PVOID pvArg1, \
      PVOID pvArg2, \
      DWORD dwFlags), \
     (pvArg1, pvArg2, dwFlags)) \
PROC(FT_Close, \
     (FT_HANDLE ftHandle), \
     (ftHandle)) \
PROC(FT_Read, \
     (FT_HANDLE ftHandle, \
      LPVOID lpBuffer, \
      DWORD dwBytesToRead, \
      LPDWORD lpdwBytesReturned), \
     (ftHandle, lpBuffer, dwBytesToRead, lpdwBytesReturned)) \
PROC(FT_Write, \
     (FT_HANDLE ftHandle, \
      LPVOID lpBuffer, \
      DWORD dwBytesToWrite, \
      LPDWORD lpdwBytesWritten), \
     (ftHandle, lpBuffer, dwBytesToWrite, lpdwBytesWritten)) \
PROC(FT_ResetDevice, \
     (FT_HANDLE ftHandle), \
     (ftHandle)) \
PROC(FT_SetTimeouts, \
     (FT_HANDLE ftHandle, \
      DWORD dwReadTimeout, \
      DWORD dwWriteTimeout), \
     (ftHandle, dwReadTimeout, dwWriteTimeout))

// enumerate procedure IDs
#define PROC(NAME, PARAMTYPES, PARAMLIST) ID_##NAME, 
enum { USB_PROCS FT_MAX };
#undef PROC


static void *ftFunctions[FT_MAX]; // pointers to USB procedures
static HMODULE usbDll = NULL;     // handle of USB DLL


// enumerate procedure names
#define PROC(NAME, PARAMTYPES, PARAMLIST) #NAME ,
static const char *usbProcNames[] = { USB_PROCS };
#undef PROC

// declare procedure types
#define PROC(NAME, PARAMTYPES, PARAMLIST) \
  typedef FTD2XX_API FT_STATUS (WINAPI *TYPE_##NAME) PARAMTYPES;
USB_PROCS
#undef PROC


//generate wrapper function of each procedure
#define PROC(NAME, PARAMTYPES, PARAMLIST) \
  FT_STATUS USB_##NAME PARAMTYPES { \
    return (*(TYPE_##NAME)(ftFunctions[ID_##NAME])) PARAMLIST; \
  }
USB_PROCS
#undef PROC

// generate each procedure's wrapper function with error check.
#define PROC(NAME, PARAMTYPES, PARAMLIST) \
  FT_STATUS EC_USB_##NAME PARAMTYPES { \
    FT_STATUS ftStatus = USB_##NAME PARAMLIST; \
    if(!FT_SUCCESS(ftStatus)){ \
      fprintf(stderr, "Error : " #NAME "failed : %s\n", \
              getFtErrorMessage(ftStatus)); \
      exit(1); \
    } \
    return ftStatus; \
  }
USB_PROCS
#undef PROC


//-----------------------------------------------------------------------------
// returns an error message coressponds to the ftStatus
//-----------------------------------------------------------------------------
static const char* getFtErrorMessage(FT_STATUS ftStatus)
{
  static const char* ftErrorMessages[] = {
    "No error",
    "Invalid Handle",
    "Device not found",
    "Device not opened",
    "I/O error",
    "Insufficient resouces",
    "Invalid parameter",
  };
  
  if(((UINT)ftStatus) >= N_ELEMS(ftErrorMessages)){
    return "Unknown Error";
  } else {
    return ftErrorMessages[ftStatus];
  } 
}

//-----------------------------------------------------------------------------
// setupDll
//-----------------------------------------------------------------------------
static void setupDll()
{
  unsigned u;
  const char* filename = "Ftd2xx.dll";
  usbDll = LoadLibrary(filename);
  if(usbDll == NULL) {
    fprintf(stderr, "Error: Can't Load %s\n", filename);
    exit(1);
  }
	
  for(u = 0; u < N_ELEMS(usbProcNames); u++){
    void *proc = (void*)GetProcAddress(usbDll, usbProcNames[u]);
    if(proc == NULL){
      fprintf(stderr, "Error : Can't Find %s\n", usbProcNames[u]);
      exit(1);
    }
    ftFunctions[u] = proc;
  }
}

//-----------------------------------------------------------------------------
// cleanupDll
//-----------------------------------------------------------------------------
static void cleanupDll()
{
  if(usbDll != NULL){
    FreeLibrary(usbDll);
    usbDll = NULL;
  }
}

//*****************************************************************************
// Applicationd dependent part.
//*****************************************************************************

const  FT_HANDLE FT_NULL_HANDLE = (FT_HANDLE)-1;
static FT_HANDLE ftHandle = FT_NULL_HANDLE;

//-----------------------------------------------------------------------------
// cleanApp
//-----------------------------------------------------------------------------
static void cleanApp()
{
  if(ftHandle != FT_NULL_HANDLE){
    USB_FT_ResetDevice(ftHandle);
    USB_FT_Close(ftHandle);
    ftHandle = FT_NULL_HANDLE;
  }
  cleanupDll();
}

//-----------------------------------------------------------------------------
// initUSB
//-----------------------------------------------------------------------------
static void initUSB()
{
  setupDll();
  atexit(cleanApp);
}

//-----------------------------------------------------------------------------
// main
//-----------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  const unsigned SRAM_SIZE = 4*1024*1024;
  DWORD nBytes;
  char buffer[4096];
  int c,n;
  //const char *testfile = "test.dat";

  // Open output file for test
	//FILE* fpr = fopen(testfile,"wb");
    //if(fpr == NULL){
    //    fprintf(stderr, "Error : Can't open %s\n",testfile);
    //	exit(1);
    //}

  // Open input file & Prepare Data
  const char *fileName = (argc < 2) ? "sramw.dat" : argv[1];
  FILE* fp = fopen(fileName, "rb");
  if(fp == NULL){
    fprintf(stderr, "Error : Can't open %s\n", fileName);
    exit(1);
  }


  
  //printf("Write Data : %d bytes\n",n);

  // Open USB Device
  initUSB();
  EC_USB_FT_Open(0, &ftHandle);
  fprintf(stderr, "Open OK, handle is %p\n", ftHandle);

  // Set timeout
  EC_USB_FT_SetTimeouts(ftHandle, 1000, 1000); // timeout = 1000ms

  // Reset and wait for the USB device to settle down
  EC_USB_FT_ResetDevice(ftHandle);
  Sleep(1000);  // wait for 1000ms

  // Write a byte to trigger the data transfer from XC2V1000 circuit
  EC_USB_FT_Write(ftHandle, buffer, 1, &nBytes);
  
  // Read a byte to trigger
  //EC_USB_FT_Read(ftHandle, buffer, 1, &nBytes);
      
  printf("Write Start.\n");

  // Write a byte to trigger the data transfer from XC2V1000 circuit
//  EC_USB_FT_Write(ftHandle, buffer, 1, &nBytes);

  // write 4M bytes

  DWORD totalSize = 0;
  
  while(totalSize < SRAM_SIZE){

	fread(buffer,1,sizeof(buffer),fp);

    DWORD remain = SRAM_SIZE - totalSize;
    DWORD toWrite = (remain > sizeof(buffer)) ? sizeof(buffer) : remain;
    
    fprintf(stderr, "about to write %lu bytes...", toWrite);
    EC_USB_FT_Write(ftHandle, buffer, toWrite, &nBytes);
    //fwrite(buffer, 1, nBytes, fpr); // for test
    totalSize += nBytes;
    fprintf(stderr, "Write %lu bytes(total %lu bytes)\r", nBytes, totalSize);
  }

  fclose(fp);
  //fclose(fpr);


  fprintf(stderr, "\ndone.\n");

  // try to read more data to assume that no redundant byte is sent.
//  EC_USB_FT_Read(ftHandle, buffer, sizeof(buffer), &nBytes);
//  fprintf(stderr, "extra data size = %lu\n", nBytes);
  return 0;
}
